@@ -13,7 +13,6 @@ use Mojo::JSON 'to_json', 'from_json', 'encode_json', 'decode_json';
 use Mojo::WebSocket 'WS_PING';
 use Mojo::IOLoop;
 use RxPerl::Mojo ':all';
-use Crypt::PRNG 'rand';
 use Safe::Isa;
 use Syntax::Keyword::Try;
 use Struct::Diff 'diff';
@@ -27,7 +26,7 @@ use Encode 'encode_utf8';
 
 use experimental 'postderef';
 
-our $VERSION = "v0.0.9";
+our $VERSION = "v0.0.10";
 
 use constant {
     DEFAULT_HEARTBEAT_INTERVAL        => 5,
@@ -73,12 +72,14 @@ sub register ($self, $app, $config) {
     my $last_heartbeat_dt;
 
     my sub calculate_hb_expiry_dur () {
-        return $heartbeat_interval + max($heartbeat_interval * 0.5, 10);
+        state $hb_e_d = $heartbeat_interval + max($heartbeat_interval * 0.5, 10);
     }
 
     # on workers startup:
     my $ping_o;
     Mojo::IOLoop->next_tick(sub {
+        srand();
+
         # ping every 15 seconds
         $ping_o = rx_defer(sub {
             rx_timer(rand($ping_interval), $ping_interval);
@@ -98,40 +99,52 @@ sub register ($self, $app, $config) {
 
         # remove any heartbeat record
         $graceful_stop_o->subscribe(sub {
-            $app->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
-                delete $state->{$worker_uuid};
-
-                return undef, $state;
-            }, no_ban => 1);
+            # don't prevent other callbacks if this dies
+            eval {
+                $app->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
+                    delete $state->{$worker_uuid};
+                    return undef, $state;
+                }, no_ban => 1);
+            };
         });
 
-        # register heartbeat
-        $app->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
-            $last_heartbeat_dt = int get_time;
-            $state->{$worker_uuid} = {
-                uuid      => $worker_uuid,
-                pid       => $$,
-                heartbeat => $last_heartbeat_dt,
-                hostname  => hostname,
-                banned    => false,
-            };
-            return undef, $state;
-        }, no_ban => 1);
+        # register heartbeat at start, or gracefully stop
+        try {
+            $app->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
+                $last_heartbeat_dt = int get_time;
+                $state->{$worker_uuid} = {
+                    uuid      => $worker_uuid,
+                    pid       => $$,
+                    heartbeat => $last_heartbeat_dt,
+                    hostname  => hostname,
+                    banned    => false,
+                };
+                return undef, $state;
+            }, no_ban => 1);
+        } catch {
+            Mojo::IOLoop->stop_gracefully;
+        };
+
+        # register heartbeat every X seconds, or gracefully stop
         rx_merge(
             rx_timer(rand($heartbeat_interval), $heartbeat_interval),
         )->pipe(
             op_take_until($graceful_stop_o),
         )->subscribe(sub {
-            $app->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
-                my $me = $state->{$worker_uuid};
-                if (! $me or $me->{banned}) {
-                    Mojo::IOLoop->stop_gracefully;
-                    return undef, undef;
-                }
-                $last_heartbeat_dt = int get_time;
-                $me->{heartbeat} = $last_heartbeat_dt;
-                return undef, $state;
-            }, no_ban => 1);
+            try {
+                $app->$bs_prefix->lock_state($WORKERS_CHANNEL, sub ($state) {
+                    my $me = $state->{$worker_uuid};
+                    if (! $me or $me->{banned}) {
+                        Mojo::IOLoop->stop_gracefully;
+                        return undef, undef;
+                    }
+                    $last_heartbeat_dt = int get_time;
+                    $me->{heartbeat} = $last_heartbeat_dt;
+                    return undef, $state;
+                }, no_ban => 1);
+            } catch {
+                Mojo::IOLoop->stop_gracefully;
+            };
         });
 
         $pubsub_connected_o->subscribe(sub {
@@ -164,8 +177,22 @@ sub register ($self, $app, $config) {
             ),
         )->subscribe(sub { Mojo::IOLoop->stop_gracefully });
 
-        # detect disconnects by opening a permanent dummy listen
-        $pg->pubsub->listen('!system' => sub ($pubsub, $payload) {});
+        # if no notifications from _bs:workers in a while, stop gracefully
+        rx_observable->new(sub ($subscriber) {
+            try {
+                $pg->pubsub->listen($WORKERS_CHANNEL => sub ($pubsub, $payload) {
+                    $subscriber->next();
+                });
+            } catch {
+                Mojo::IOLoop->stop_gracefully;
+            };
+        })->pipe(
+            op_switch_map(sub {
+                return rx_timer(calculate_hb_expiry_dur());
+            }),
+        )->subscribe(sub {
+            Mojo::IOLoop->stop_gracefully;
+        });
     });
 
     $app->helper("$bs_prefix.add_action", sub ($c, $channel_name, $action_name, $action_sub) {
